@@ -17,13 +17,12 @@
 """
 
 
-import json
 import socket
-import urllib.request
 import urllib.parse
-import http.cookiejar
+import urllib.error
 import http.client
-import time
+import http.cookiejar
+import json
 
 import pytz
 
@@ -32,7 +31,7 @@ from ...optconf import SecretOption
 from ...optconf.converters import as_string_or_none
 
 from ... import tfile
-from ... import sockshandler
+from ... import web
 
 from .. import BasePlugin
 from .. import BaseExtension
@@ -51,75 +50,13 @@ class LogicError(FetcherError):
     pass
 
 
-class TypicalNetworkError(FetcherError):
-    def __init__(self, err):
+class NetworkError(FetcherError):
+    def __init__(self, sub):
         super().__init__()
-        self.err = err
+        self._sub = sub
 
     def __str__(self):
-        return "{}: {}".format(type(self.err).__name__, str(self.err))
-
-
-# =====
-def select_fetcher(torrent, fetchers):
-    for fetcher in fetchers:
-        if fetcher.is_matched_for(torrent):
-            return fetcher
-    return None
-
-
-# =====
-def build_opener(proxy_url=None, cookie_jar=None):
-    handlers = []
-
-    if proxy_url is not None:
-        scheme = (urllib.parse.urlparse(proxy_url).scheme or "").lower()
-        if scheme == "http":
-            proxies = dict.fromkeys(("http", "https"), proxy_url)
-            handlers.append(urllib.request.ProxyHandler(proxies))
-        elif scheme in ("socks4", "socks5"):
-            handlers.append(sockshandler.SocksHandler(proxy_url=proxy_url))
-        else:
-            raise RuntimeError("Invalid proxy protocol: {}".format(scheme))
-
-    if cookie_jar is not None:
-        handlers.append(urllib.request.HTTPCookieProcessor(cookie_jar))
-
-    return urllib.request.build_opener(*handlers)
-
-
-def read_url(
-    opener,
-    url,
-    data=None,
-    headers=None,
-    timeout=10,
-    retries=10,
-    sleep_time=1,
-    retry_codes=(500, 502, 503),
-    retry_timeout=True,
-):
-    while True:
-        try:
-            request = urllib.request.Request(url, data, (headers or {}))
-            return opener.open(request, timeout=timeout).read()
-        except socket.timeout as err:
-            if retries == 0 or not retry_timeout:
-                raise TypicalNetworkError(err)
-        except urllib.error.HTTPError as err:
-            if retries == 0 or err.code not in retry_codes:
-                raise TypicalNetworkError(err)
-        except urllib.error.URLError as err:
-            if isinstance(err.reason, str) and "timed out" in err.reason:
-                if retries == 0 or not retry_timeout:
-                    raise TypicalNetworkError(err)
-            else:
-                raise
-        except (http.client.IncompleteRead, http.client.BadStatusLine, ConnectionResetError) as err:
-            if retries == 0:
-                raise TypicalNetworkError(err)
-        time.sleep(sleep_time)
-        retries -= 1
+        return "{}: {}".format(type(self._sub).__name__, str(self._sub))
 
 
 # =====
@@ -129,14 +66,13 @@ def _assert(exception, arg, msg=""):
 
 
 class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
-    def __init__(self, url_retries, url_sleep_time, timeout, user_agent, client_agent, proxy_url,
+    def __init__(self, timeout, retries, retries_sleep, user_agent, proxy_url,
                  check_version, check_fingerprint, **_):
 
-        self._url_retries = url_retries
-        self._url_sleep_time = url_sleep_time
         self._timeout = timeout
+        self._retries = retries
+        self._retries_sleep = retries_sleep
         self._user_agent = user_agent
-        self._client_agent = client_agent
         self._proxy_url = proxy_url
         self._check_version = check_version
         self._check_fingerprint = check_fingerprint
@@ -157,12 +93,11 @@ class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
     @classmethod
     def get_options(cls):
         return {
-            "url_retries":       Option(default=20, help="The number of retries to handle tracker-specific HTTP errors"),
-            "url_sleep_time":    Option(default=1.0, help="Sleep interval between the retries"),
             "timeout":           Option(default=10.0, type=float, help="Timeout for HTTP client"),
-            "user_agent":        Option(default="Mozilla/5.0", help="User-agent for site"),
-            "client_agent":      Option(default="rtorrent/0.9.2/0.13.2", help="User-agent for tracker"),
-            "proxy_url":         Option(default=None, type=as_string_or_none, help="The URL of the HTTP proxy"),
+            "retries":           Option(default=20, help="The number of retries to handle tracker-specific HTTP errors"),
+            "retries_sleep":     Option(default=1.0, help="Sleep interval between failed retries"),
+            "user_agent":        Option(default="Mozilla/5.0", help="User-Agent for site"),
+            "proxy_url":         Option(default=None, type=as_string_or_none, help="URL of HTTP/SOCKS4/SOCKS5 proxy"),
             "check_fingerprint": Option(default=True, help="Check the site fingerprint"),
             "check_version":     Option(default=True, help="Check the fetcher version from GitHub"),
         }
@@ -178,28 +113,41 @@ class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
     def _init_opener(self, with_cookies):
         if with_cookies:
             self._cookie_jar = http.cookiejar.CookieJar()
-            self._opener = build_opener(self._proxy_url, self._cookie_jar)
+            self._opener = web.build_opener(self._proxy_url, self._cookie_jar)
         else:
-            self._opener = build_opener(self._proxy_url)
+            self._opener = web.build_opener(self._proxy_url)
 
     def _build_opener(self):
-        return build_opener(self._proxy_url)
+        return web.build_opener(self._proxy_url)
 
-    def _read_url(self, url, data=None, headers=None, opener=None):
+    def _read_url(self, *args, **kwargs):
+        try:
+            return self._read_url_nofe(*args, **kwargs)
+        except (
+            socket.timeout,
+            urllib.error.HTTPError,
+            urllib.error.URLError,
+            http.client.IncompleteRead,
+            http.client.BadStatusLine,
+            ConnectionResetError,
+        ) as err:
+            raise NetworkError(err)
+
+    def _read_url_nofe(self, url, data=None, headers=None, opener=None):
         opener = (opener or self._opener)
         assert opener is not None
 
         headers = (headers or {})
         headers.setdefault("User-Agent", self._user_agent)
 
-        return read_url(
+        return web.read_url(
             opener=opener,
             url=url,
             data=data,
             headers=headers,
             timeout=self._timeout,
-            retries=self._url_retries,
-            sleep_time=self._url_sleep_time,
+            retries=self._retries,
+            retries_sleep=self._retries_sleep,
             retry_codes=self._retry_codes,
         )
 
@@ -223,7 +171,7 @@ class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
 
     def test(self):
         if self._check_fingerprint or self._check_version:
-            opener = build_opener(self._proxy_url)
+            opener = web.build_opener(self._proxy_url)
             info = self._get_upstream_info(opener)
         if self._check_fingerprint:
             self._test_fingerprint(info["fingerprint"], opener)
@@ -231,10 +179,15 @@ class BaseFetcher(BasePlugin):  # pylint: disable=too-many-instance-attributes
             self._test_version(info["version"])
 
     def _get_upstream_info(self, opener):
-        return json.loads(self._read_url(
-            url="https://raw.githubusercontent.com/mdevaev/emonoda/master/fetchers.json",
-            opener=opener,
-        ).decode("utf-8")).get(self.get_name(), self._get_local_info())
+        try:
+            return json.loads(self._read_url_nofe(
+                url="https://raw.githubusercontent.com/mdevaev/emonoda/master/fetchers/{}.json".format(self.get_name()),
+                opener=opener,
+            ).decode("utf-8"))
+        except urllib.error.HTTPError as err:
+            if err.code == 404:
+                return self._get_local_info()
+            raise
 
     @classmethod
     def _get_local_info(cls):
@@ -282,6 +235,26 @@ class WithCaptcha(BaseExtension):
         self._captcha_decoder = captcha_decoder
 
 
+class WithScrape(BaseExtension):
+    def __init__(self, client_agent, **_):
+        self._client_agent = client_agent
+
+    @classmethod
+    def get_options(cls):
+        return {
+            "client_agent": Option(default="rtorrent/0.9.2/0.13.2", help="User-Agent for tracker"),
+        }
+
+    def _is_torrent_registered(self, base_scrape_url, torrent):
+        # https://wiki.theory.org/BitTorrentSpecification#Tracker_.27scrape.27_Convention
+        self._assert_match(torrent)  # pylint: disable=no-member
+        data = self._read_url(  # pylint: disable=no-member
+            url=urllib.parse.urljoin(base_scrape_url, "scrape.php?info_hash={}".format(torrent.get_scrape_hash())),
+            headers={"User-Agent": self._client_agent},
+        )
+        return (len(tfile.decode_data(data).get("files", {})) == 0)
+
+
 class WithTime(BaseExtension):
     def __init__(self, timezone, **_):
         self._default_timezone = timezone
